/* digiconf.c
 *:set ts=4   // Formatted for tabstops every 4 characters
 *
 * Present a list of adapters supported by Digi's "epca" Linux driver,
 * and lead the user through a sufficiently detailed interview to fully
 * specify configuration options for all the adapters they wish to
 * configure.  Then write out the "epcaconfig.h" file used by the driver
 * itself (at compile time).
 *
 * Supported adapters and their characteristics and options are specified
 * in the "digiconf.h", which makes coordinates definitions with the driver
 * by referencing several driver header files (epca.h and digiPCI.h).
 *
 * Version 0.1: January 12, 2000: Peter Berger and Chad Schwartz
 * Version 0.2: January 14, 2000: PEB/CWS
 *     -- Comment out references to PCI boards in epcaconfig.h since
 *        the current driver (v1.4.2) autoprobes for these.
 *     -- Converted get_int()'s "other" parameter from a charatcter to a
 *        string
 *     -- Fixed the "number of ports" query so it only asks the user if
 *        there's more than one possible choice.
 *
 * Version 0.3: January 17, 2000: CWS
 *     -- Why ask for ALTPIN on PCI boards, only to have it NOT work for
 *	  the user?  Now, we only ask for ALTPIN on ISA serial boards. (We
 *	  can change this in the future, once PCI boards are NOT autoprobed.)
 *
 * Version 0.3.1: January 17, 2000: CWS
 *     -- We now know when a C/X board has been installed.  When a user 
 *        selects a C/X board (any type) digiConf knows to go ahead and
 *  	  run cxconf for the user. As of now, we just run it from the
 *        default path.  Later, we may want to call it DIRECTLY from where
 *        we know it will be. 
 * 
 * Version 0.4: January 19, 2000: CWS
 *     -- Oops! We weren't keeping track of the fact that when we have
 *	  PCI boards, we do NOT add those into NBDEVS and NUMCARDS.  If
 *	  we do, VERY BAD THINGS HAPPEN.
 *     -- We now run /etc/digi/buildPCI directly from here - we build the
 *	  number of devices specified on the fly, and then we tell the 
 *	  user what their tty devices are.
 *
 *
 * Version 0.5: February 22, 2000: CWS
 *     -- Added more portcounts in Xem/CX configuration/setup
 *     -- We weren't tracking portcounts on ISA boards properly, thus when
 *        the user went to install their ISA digiboard with a certain port
 *        count, it gave them a DIFFERENT number of ports, than they asked
 *        for. Fixed.
 *
 * Version 0.5.1: March 8, 2000: CWS
 *     -- We were using incorrect logic in determining which digiConf
 *        board type to use, when writing out the epcaconfig.h file.
 *        the problem originally manifested itself when a user tried
 *        to install a C/X EISA board. (The driver would think the
 *        board was PCI - because it was looking at EISACX type 9 as
 *        defined in epca.h - but it really needed to look at digiConf
 *        type 8, C/X EISA configuration array.)
 *     -- Fixed potential segfault bug while configuring PCI DigiBoards.
 *        
 * Version 0.6: February 25, 2000: CWS
 *     -- Added support for the new DPA compatibility information in
 *        generation of epcaconfig.h.  We now pass the DPA type of board
 *	  in properly.
 *     -- Removed support for Xeve/Xe64k/Xem-EISA/CX-EISA boards, as these
 *	  boards either A) don't apply anymore, or B) are obsoleted.
 *
 * Version 0.6.5: March 2, 2000: CWS
 *     -- Changed path from which to run cxconf
 *
 * Version 0.6.6: March 24, 2000: CWS
 *     -- Added EPC/X Jupiter board support
 *     -- Added "Just Press Enter" support in get_int(), which enables
 *        us to paginate the portcount question properly. 
 *
 * Version 0.6.7: September 6, 2000: CWS
 *     -- Fixed up a bunch of errors that showed up when compiling w/ GCC
 *	  set to -Wall
 *     -- Made digiConf spit out /etc/digiepca/modcmdline.sh - which contains
 *	  the module load cmdline options for epca.o.  This is now the default
 *	  behavior.
 *     -- Added default modcmdline.sh - that gets installed with the source
 * 	  RPM package. it issues an error message - that the driver hasn't
 *        been configured yet, and exits the shell with -1 so that rc.epca
 *        can detect the error, and not attempt to go any further.
 *     -- Added --epcaconfig-mode flag that makes the driver generate the 
 *	  old-style 'epcaconfig.h' file in the current directory. This
 *	  behavior is obsolete, and is no longer needed for current methods
 *	  of operation.
 *  
 * Version 0.6.8: April 24, 2001: JAR
 *     -- Changed name of modcmdline.sh to modcmdline
 *
 */

#define VERSION "0.6.8"
char version[]=VERSION;

#define MAXBUFSIZE 1024

typedef unsigned char unchar;
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

#include <fcntl.h>
#include <termio.h>

#include "rascon.h"
#include "epca.h"
#include "digiPCI.h"

#include "digiConf.h"

int cxthere=0;  	/* Do we have any C/X boards? */
int epcxthere=0;	/* Do we have any EPC/X boards? */

char *outfname = NULL;  /* Lets just shove argv[1] into this so we have */
			/* access to it within all functions, but we    */
			/* don't have to go through the crap of passing */
			/* it down to every function... We'll just      */
			/* strdup() it in here... CWS                   */

char lbuf[50];

unsigned int num_candidates=sizeof(adapters)/sizeof(struct adapter_struct) - 1;


int modulecmdlinemode = 1;

char *busname( unsigned int bus ) {
	char *ret;

	if( bus == ISA )
		ret = "ISA";
	else if( bus == EISA )
		ret = "EISA";
	else if( bus == PCI )
		ret = "PCI";
	else
		ret = "Unknown";
		
	return( ret );
}

void pressenter(char *msg) {
	if(msg) 
		printf("%s",msg);
	fgets(lbuf, sizeof(lbuf), stdin);
	
	if(msg) printf("\n");
}

#define GETINT_MAXLEN 10
int get_int( int pressenter, int min, int max, char *other, char *prompt, ... ) {
	char resp[GETINT_MAXLEN];
	int ret = 0, ask_again;
	va_list args;
	
	va_start( args, prompt );

	do {
		ask_again = 0;
  		vprintf( prompt, args );
		printf( " (%d-%d)", min, max );
		if( fgets( resp, GETINT_MAXLEN, stdin ) == NULL) {
				ask_again = 1;
				printf( "\n" );
		} else if( *other && strchr(other, resp[0]) ) {
			ret = -resp[0];
			ask_again = 0;
		} else if( ((ret=atoi(resp)) < min) || (ret > max) ) {
			if(!pressenter)
				ask_again = 1;
			else
				return(-1);
		}
	} while( ask_again );

	va_end( args );
	return( ret );
}

#define	YESNO_MAXLEN 10
int yesno( char *prompt, ... ) {
	va_list args;
	int ret = 0, ask_again;
	char resp[YESNO_MAXLEN];

	va_start( args, prompt );

	do {
		ask_again = 0;
  		vprintf( prompt, args );
		printf( " ('y' or 'n')" );
		if( fgets( resp, YESNO_MAXLEN, stdin ) == NULL) {
			ask_again = 1;
			printf( "\n" );
		} else if( toupper(resp[0]) == 'Y' )
			ret = 1;
		else if( toupper(resp[0]) == 'N' )
			ret = 0;
		else
			ask_again = 1;
	} while( ask_again );
	va_end( args );

	return( ret );
}

void show_boards() {
	int i;

	for( i=0; i<num_candidates; ++i ) {
		printf( "     %d: %s\n",
			i+1, btypes[adapters[i].boardid].desc );
	}

}

unsigned int choose_board( unsigned int i, unsigned int nadapters ) {
	int bd;
	
	while( (bd=get_int(0, 1, num_candidates, "Ll",
		"Board #%d. What type of board is this? ('L' for list)",
		i+1 )) < 0 ) {
				if( toupper( -bd ) == 'L' )
					show_boards();
	}
	--bd;	// Internally we count from 0; user counts from 1

	printf( "Great! You've selected to install a %s board!\n",
		btypes[adapters[bd].boardid].desc );
	
	if(adapters[bd].boardid == PCICX ||
		adapters[bd].boardid == ISACX) {
		
		/* Okay. We've got a C/X board. Lets remember to run cxconf */
		/* when we're done..CWS					    */
		cxthere++;
	}	
	
	if(adapters[bd].boardid == PCIEPCJ) {
		/* Found an EPC/X board. Lets remember to run epcxconf 	*/
		/* when we're done..CWS					*/
		epcxthere++;
	}
		
	return( bd );
}

unsigned int get_numports( unsigned bd_index, unsigned int bd ) {
	int i, nchoices, ret, numopts, j=0;

	// How many port count choices are available for this adapter?
	nchoices=0;
	while( adapters[bd].nports_array[nchoices] )
		nchoices++;

	if( nchoices < 2 ) {
		printf( "This digiBoard has %d ports\n",
			(ret=adapters[bd].nports_array[0]) );
	} else {
		printf( "How many ports does this digiBoard have?  Possible values:\n" );
		
		for(i=0; adapters[bd].nports_array[i]; i++) {
			numopts++;
		}
		
		for( i=0; adapters[bd].nports_array[i]; ++i ) {
			if(j==20) {
				ret=get_int(1, 1, i, "", "Board #%d. How many ports? ( Press [ENTER] for more ) ",
					bd_index+1 );
				if(ret == -1) {
					j=0;
				} else {
					return(ret);
				}
			}
			j++;

			printf( "     %d: %d\n", i+1, adapters[bd].nports_array[i] );
		}
			
		ret=get_int( 0, 1, i, "", "Board #%d. How many ports? ", bd_index+1 ); 
	}

	return( ret );
}
		

/* Get the configuration info for this adapter, stuff the values in the
 * configuration array, and return the number of ports on this adapter.
 */
unsigned int config_board( unsigned bd_index, unsigned int bd ) {

	unsigned int i, io=0, mem=0, nports, altpin, nchoices, ret=0;

	if( adapters[bd].bus == ISA || adapters[bd].bus == EISA ) {
		printf( "This board needs an I/O address.  Possible I/O addresses:\n" );
		for( i=0; adapters[bd].io_array[i]; ++i ) {
			printf( "     %d: %xx\n", i+1, adapters[bd].io_array[i] );
		}
		io=get_int(0, 1, i, "",
			"Board #%d. What I/O address would you like to use?", bd_index+1 ); 

		printf(
"Great! You now need to select a SHARED MEMORY ADDRESS for your %s %s\n"\
"DigiBoard.  Selecting a shared memory address that is not in use by other\n"\
"components of your system is VERY IMPORTANT Here.  For example, if you have\n"\
"a device on your system that uses 0xD0000, make SURE you do NOT select that\n"\
"address for your DigiBoard.\n",
busname(adapters[bd].bus), btypes[adapters[bd].boardid].desc );
		printf( "Possible shared memory addresses:\n" );
		for( i=0; adapters[bd].mem_array[i]; ++i ) {
			printf( "     %d: %xx\n", i+1, adapters[bd].mem_array[i] );
		}
		mem=get_int( 0, 1, i, "",
			"Board #%d. What shared memory address would you like to use?", bd_index+1 ); 

	} else {
		printf(
		"Memory addresses will be read from the PCI card itself.\n" );
	}

	nports=get_numports( bd_index, bd );

	altpin=yesno( "Do you want to set Altpin on this board? " );
		
	// Store the values for this board (later we'll write them to epcaconfig.h)
	epcaconf[bd_index].status = ENABLED;
	epcaconf[bd_index].type = adapters[bd].boardid;
	epcaconf[bd_index].altpin = altpin? ON : OFF; 

	if(adapters[bd].bus != PCI) {
		epcaconf[bd_index].port = adapters[bd].io_array[io-1];
		epcaconf[bd_index].membase = adapters[bd].mem_array[mem-1];
	}
	
	nchoices = 0;
		
	while(adapters[bd].nports_array[nchoices] )
	nchoices++;
		
	if(nchoices < 2) {
		ret = 0;
	} else {
		ret = nports - 1;
	}
		
		
	epcaconf[bd_index].numports = adapters[bd].nports_array[ret];
	         	
	if( adapters[bd].bus == PCI ) {
		epcaconf[bd_index].port = 0;
		epcaconf[bd_index].membase = 0;
	}

	/* Return the preferred value, whichever it is.. */
	
	return(adapters[bd].nports_array[ret] );
}

void write_epcamodulecmdlineconfig( int nadapters, int numports ) {
	FILE *outf;
	int i;
	char mod_type[MAXBUFSIZE];
	char mod_altpin[MAXBUFSIZE];
	char mod_numports[MAXBUFSIZE];
	char mod_port[MAXBUFSIZE];
	char mod_membase[MAXBUFSIZE];
	
	if( (outf = fopen( outfname, "w" )) == NULL ) {
		printf("Cannot open output file: %s\n", outfname );
		return;
	}
	chmod(outfname, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	
	/* set up type */
	sprintf(mod_type, "mod_type=");
	for(i=0;i<nadapters;i++) {
		sprintf(lbuf,"%d",epcaconf[i].type);
		strcat(mod_type, lbuf);
		if(i+1 < nadapters) 
			strcat(mod_type, ",");
	}
	/* set up altpin */
	sprintf(mod_altpin, "mod_altpin=");
	for(i=0;i<nadapters;i++) {
		sprintf(lbuf, "%d",epcaconf[i].altpin);
		strcat(mod_altpin, lbuf);
		if(i+1 < nadapters)
			strcat(mod_altpin, ",");
	}
	/* set up numports */
	sprintf(mod_numports,"mod_numports=");
	for(i=0;i<nadapters;i++){
		sprintf(lbuf, "%d", epcaconf[i].numports);
		strcat(mod_numports, lbuf);
		if(i+1 < nadapters)
			strcat(mod_numports, ",");
	}
	/* set up port */
	sprintf(mod_port, "mod_port=");
	for(i=0;i<nadapters;i++){
		sprintf(lbuf, "0x%lx", epcaconf[i].port);
		strcat(mod_port,lbuf);
		if(i+1 < nadapters)
			strcat(mod_port, ",");
	}
	/* set up membase */
	sprintf(mod_membase, "mod_membase=");
	for(i=0;i<nadapters;i++){
		sprintf(lbuf, "0x%lx", epcaconf[i].membase);
		strcat(mod_membase,lbuf);
		if(i+1 < nadapters)
			strcat(mod_membase, ",");
	}
	
	
	fprintf(outf, "#!/bin/sh\n\n");
	fprintf(outf, "# %s written by digiConf version %s\n\n",
		outfname, VERSION );
	
	fprintf(outf, "insmod epca %s %s %s %s %s\n", 
		mod_type,
		mod_altpin,
		mod_numports,
		mod_port,
		mod_membase
	);
	
	fprintf(outf, "\n\nif [ ! $? -eq 0 ]; then\n");
	fprintf(outf, "		echo ""Module load failed""\n");
	fprintf(outf, "		exit -1\n");
	fprintf(outf, "fi\n");
	
	fclose(outf);
}

void write_epcaconfig( int nadapters, int numports ) {
	unsigned int bd_index;
	FILE *outf;

	if( (outf = fopen( outfname, "w" )) == NULL ) {
		printf( "Cannot open output file: %s\n", outfname );
		return;
	}

	fprintf( outf, "// %s written by digiConf version %s\n",
		outfname, VERSION );
		
	fprintf( outf, "#define NUMCARDS %d\n", nadapters);
	fprintf( outf, "#define NBDEVS %d\n\n", numports);

	fprintf( outf, "struct board_info static_boards[NUMCARDS]={\n" );

	for( bd_index=0; bd_index < nadapters; ++bd_index ) {
		
		fprintf( outf, "\t{ ENABLED, %d, %d, %d, 0x%lx, 0x%lx },\t// %s\n",
		 	epcaconf[bd_index].type,
			epcaconf[bd_index].altpin,
			epcaconf[bd_index].numports,
			epcaconf[bd_index].port,
			epcaconf[bd_index].membase,
			btypes[epcaconf[bd_index].type].desc
 		);
	}
	fprintf( outf, "};\n\n" );
	fprintf( outf, "/* DO NOT HAND EDIT THIS FILE! */\n" );
}


void argv_parse(int argc, char **argv) {
	int i;
		
	for(i=1;i<argc;i++) {
		if(!strcmp(argv[i], "--epcaconfig-mode"))
			modulecmdlinemode=0;
	}
}

int main(int argc, char *argv[]) {
	unsigned int nadapters, i, bd, nports;
	char lbuf[50];
	
	nports=0;	// Total number of ports configured so far


	argv_parse(argc, argv);
	
	if(!modulecmdlinemode)
		outfname = "epcaconfig.h";
	else
		outfname = ETCDIR"/modcmdline";
	
	
	/* Lets give 'em a couple lines.. CWS */
	printf("\n\n");

// How many adapters do they want to configure
	nadapters = get_int(0,  1, MAXBOARDS, "",
		"How many boards would you like to install?" );
	printf( "Great! we'll install %d board/s for you.\n\n", nadapters );

// Select and configure the boards
	for( i=0; i<nadapters; ++i ) {
		bd = choose_board( i, nadapters );
		nports += config_board( i, bd );
		printf( \
"Great! we have ALL the information we need, to get this digiBoard running!!\n\n"
		);
	}

	if(!modulecmdlinemode)
		write_epcaconfig( nadapters, nports );
	else
		write_epcamodulecmdlineconfig ( nadapters, nports );
		
		
	/* Okay. If C/X boards exist, lets go ahead and spawn off a */
	/* cxconf process..CWS					    */

	if(cxthere) {
		sprintf(lbuf, "%s/cxconf %d\n", SBINDIR, cxthere);
		system(lbuf);
	}
	
	if(epcxthere) {
		sprintf(lbuf, "%s/epcxconf %d\n", SBINDIR, epcxthere);
		system(lbuf);
	}
	
	return(0);
}
