/* epcxconf.c
 * Query user for EPCX concentrator topology and write the magic
 * configuration strings to cxconf.dat where digiDload expects to find them.
 *
 * Version 1.0.0: PEB: February 1999
 *
 * Version 1.0.1: PEB: March 1999
 * Added show_mode_list to display table of communications options
 *
 * Version 1.0.2: CWS: May 1999
 * Extensive functionality added, to include support for digi's EPC/X
 * line of products.  As a part of this functionality upgrade, I added
 * the ability to create both cxconf, and epcxconf, from the same source
 * file.  
 * 
 * By executing 'make cxconf', we compile the cxconf binary, that doesn't
 * have all the bells and whistles that epcxconf has.
 *
 * By executing 'make epcxconf', we compile the epcxconf binary, that has
 * all the necessary functionality compiled in, to create the 
 * epcxconf.dat file, with external concentrator functionality.
 *
 * Version 1.0.3: CWS: January 2000
 * Bugfix:  while selecting wiring codes, when we get an invalid selection,
 * we no longer exit with an error. (but rather, we just ask again.)
 * 
 * Version 1.0.4: CWS: Febuary 18 2000
 * Enhancement: Changed the way we handle lines that are not in use.  Now,
 * we actually recognize that there is a line, but it is not in use, in the 
 * C/X configuration string written to cxconf.dat.  This is so that 
 * dpa knows how to read it correctly.
 * 
 * Version 1.0.5: CWS: March 3 2000
 * Enhancement: As a result of a switch to using /etc/digiepca, we now 
 * put our resulting cxconf.dat file there.
 * Enhancement: added a get_int function, that allows us to have better
 * error control, when a user enters something incorrectly.
 * Enhancement: take number of boards as a commandline argument, rather than
 * asking user for the information.
 * Enhancement: added default options on several of the questions asked.
 * one of the most needed, is the default wiring code '14'. Now, all the
 * user has to do, is hit enter on the question, and the default answer
 * is selected.
 * Added sanity checking for concentrator portcounts question.  Now, if
 * the user doesn't use either '8' or '16' for a value, it'll complain
 * about an incorrect number of ports entered, and then ask the question
 * again.
 *
 * Version 1.0.6: CWS: July 31 2000
 * Bugfix: It was initially assumed that if a user specified a linespeed,
 * that linespeed should be carried out for all concentrators on the chain.
 * this is incorrect, seeing as though the support digi configuration - is
 * to have 1 concentrator set up w/ an alternate linespeed (or wiring type), 
 * and all other concentrators at 1.2mbit/sec (wiring code 14) to each other.
 * (As they are connected directly to each other w/ digi supplied cables.)
 * There is 1 exception to this with EPC/X boards, in that if a user enters
 * wiring code 74 (10mbit/sec) they want 10mbit/sec through the entire link.
 * Therefore, an exception is made within this code for that condition. 
 * 
 * Version 1.0.7: CWS: August 4 2000
 * Bugfix: Still didn't have C/X 1.2mbit/sec linespeed right.  Apparently,
 * due to a C/X FEP bug, we must set the last concentrator's transmit speed
 * (The speed whilst transmitting to the terminator plug) at the speed of
 * the initial adapter --> Conc #1 link.
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/* #define CWSDEBUG */
char *cxconc_hlp;
char *version = "Version 1.0.6";
FILE *fout;
char lbuf[256];
int concnum=1;
int mode=0;
char whichwiretype[64];

void stripcrlf(char *s) {
	while(*s && index("\n\r", s[strlen(s)-1]))
	s[strlen(s)-1] = 0;
}

int strisalldigit(char *s) {
	
	while(*s) {
		if(!isdigit(*s)) 
			return(0);

		s++;
	}
	return(1);
}

int get_int(int deflt, char *other, char *question, ...) {
	int tmp;
	int askagain;
	va_list args;
	va_start (args, question);

	do {
		askagain = 0;

		if(deflt != -1) {
			stripcrlf(question);
			vprintf(question, args);
			printf("[%d] : ",deflt);
		} else {
			vprintf(question, args);	
		}
		
		if( fgets(lbuf, sizeof(lbuf), stdin ) == NULL) {
			askagain = 1;
			continue;
		} else if( other && strchr(other, lbuf[0])) {
			tmp = -lbuf[0];
			return(tmp);
		}
		stripcrlf(lbuf);		/* Make sure it doesn't have a CR/LF */
		
		if(*lbuf == '\0') {
			if(deflt != -1) 
				return(deflt);
			else 
				askagain = 1;
			continue;
		}
		
		if(!strisalldigit(lbuf))        /* If it is NOT a number.. */
			askagain = 1;
		else
			tmp = atoi(lbuf);
	} while(askagain);
	va_end(args);
	return(tmp);
}

outval( int val )
{
	fprintf( fout, "%c", val );
}

#ifdef EPCXCONF
outval_or( int val )
{
	fprintf( fout, "%c", val | 0x80);

}
#endif

void show_mode_list() {
	int ret;
	
	ret = system( cxconc_hlp );
	if( ret == 127 || ret == -1 )
		printf( "Error running '%s'\n", cxconc_hlp );


}

int get_wiring_type(int card, int line) {
	int askagain=0;
	
	do {
	
		printf("\nWhat type of wiring scheme are you going to use for card %d, line %d?\n\n",card, line);
	
		printf("A) 8 Wire Direct\n");
		printf("B) 4 Wire Direct\n");
		printf("C) RS422 Sync\n");
		printf("D) RS232 Sync\n\n");
	
		printf("> ");
	
		fgets( lbuf, sizeof(lbuf), stdin);
	
		if ( toupper(lbuf[0]) == 'A' ) {
			sprintf(cxconc_hlp," more %s/cx8wiredirect.hlp",
				LIBDIR);
			return;
		}
	
		if ( toupper(lbuf[0]) == 'B' ) {
			sprintf(cxconc_hlp," more %s/cx4wiredirect.hlp",
			       LIBDIR);
			return;
		}
	
		if( toupper(lbuf[0]) == 'C' ) {
			sprintf(cxconc_hlp," more %s/cx422.hlp", LIBDIR);
			return;
		}
	
		if( toupper(lbuf[0]) == 'D' ) {
			sprintf(cxconc_hlp," more %s/cx232.hlp", LIBDIR);
			return;
		}
	
		else {
			printf("Invalid Selection!\n");
			askagain=1;
		}
	
	} while(askagain);
}
	

int get_one_comm_mode( int line, int which) {
	int tmp;
	
	tmp = get_int(14, "Ll", "    Enter the communication mode to use on line %d\n"
		             "    (Type 'L' for a list) ", line);
		       
	if(toupper(-tmp) == 'L') {
		mode = -1;
		show_mode_list();
	} else {
		mode = tmp;
	}
	
	return( mode );
}
	//if( toupper( lbuf[0] ) == 'L' ) {
	//	mode = -1;
	//	show_mode_list();
	//} else {
	//	sscanf( lbuf, "%d", &mode );
	//	}
	//return( mode );
	//}
		
	


int comm_mode( int line, int which ) {

	do {
		mode = get_one_comm_mode( line, which );
	} while (mode < 0 || mode > 76);

	return( mode );
	

}

config_one_conc( int line, int conc, int nconcs ) {
	int nports, nems, portsonem, nemscount=1;
	int ask_again;
		
#ifdef EPCXCONF
	do { 
		ask_again = 0;
		
		nports = get_int(16, NULL, "    How many ports does this EPC/CON or C/CON support? (conc #%d)\n"
			"    (NOTE: The maximum ports here, is 16) ", concnum );
		
		if(nports != 8 && nports != 16) {
			printf("\n   Invalid number of ports for concentrator %d specified!\n\n",concnum);
			ask_again = 1;
		}
	} while (ask_again);
	
#else
	do {
		ask_again = 0;
		
		nports = get_int(16, NULL, "    How many ports does this C/CON support? (conc #%d)\n"
			"    (NOTE: The maximum ports here, is 16) ", concnum);
		if(nports != 8 && nports != 16) {
			printf("\n    Invalid number of ports for concentrator %d specified!\n\n",concnum);
			ask_again = 1;
		}
	} while(ask_again);
	
#endif	
	concnum=concnum++;

#ifdef EPCXCONF
	
	nems = get_int(0, NULL, "    How many additional Ports/EM Modules do you have on this concentrator?\n"
			     "    (NOTE: Answer '0', if this is a C/CON) ");
	
	if(nems == 0) {
#ifdef CWSDEBUG
	printf( "skipping Ports/EM Config!\n");    
#endif
	}
	
	else {
	while(nemscount != nems+1) {
	
		portsonem = get_int(16, NULL, "    How many ports on Ports/EM #%d? ",nemscount);
		outval_or( portsonem );
		nemscount=nemscount++;
		}
	}
#endif
 			
	outval( nports );
	
	// outval( mode );

	/* Assume that from the first conc to the 2nd conc, the default mode */
	/* will be 14							     */

#ifdef EPCXCONF
	/* Make an exception for EPC/X boards - if the user specifies wiring */
	/* code 74, it means we want wiring code 74 throughout the entire    */
	/* chain.  This is why users BUY EPC/X's - so they can have the	     */
	/* extra bandwidth						     */
	
	if(mode == 74) 
		outval(74);	
	else
#endif
	
	/* Due to a really large FEP/OS firmware bug - that apparently has existed */
	/* since its inception, we must make sure that the wiring code at the end  */
	/* of the chain matches the one initially given to FEP/OS for the initial  */
	/* Adapter -> Conc 1 link						   */
	
	if( (conc+1) - nconcs == 0) {
		outval(mode);
	} else
		outval(14);
		
}

int config_one_line( int card, int line ) {
	int conc, nconcs;
	
#ifdef EPCXCONF	
	nconcs = get_int(-1, NULL, "How many EPC/CONs and/or C/CON's are connected to card %d, line %d? ",
	card+1, line );
#else
	nconcs = get_int(-1, NULL, "How many C/CON's are connected to card %d,  line %d? ",
	card+1, line);
#endif
	conc = 0;
	if(nconcs==0) {
		return(-1);

	} else {
	get_wiring_type( card+1, line);
	}

	if( line > 1 && nconcs > 0 )
		outval( 0 );	// Mark end of previous line
	
	while( conc<nconcs ) {
		if( conc == 0 ) {
			mode = comm_mode( line, conc );
			outval( mode );
		}
		
		config_one_conc( line, conc, nconcs);
		
		++conc;
	}

	return(0);
}
config_one_card( int card ) {
	int err;
	
	err = config_one_line( card, 1 );
	if(err == -1)
		outval(0x0e);
		
	err = config_one_line( card, 2 );
	if(err == -1) {
		outval(0x0);
		outval(0x0e);
	}
}

config(int args) {
	int	ncards = 0, card = 0;


#ifdef EPCXCONF
	if(args) {
		printf("Installing support for %d EPC/X card%s\n",
		args, args == 1 ? " " : "s");
		ncards = args;
	} else
		ncards = get_int(-1, NULL, "How many EPCX cards do you have? " );
#else
	if(args) {
		printf("Installing support for %d C/X card%s\n",
		args, args == 1 ? " " : "s");
		ncards = args;
	} else
		ncards = get_int(-1, NULL, "How many C/X cards do you have? " );
#endif
	
	//fgets( lbuf, sizeof(lbuf), stdin );
	//sscanf( lbuf, "%d", &ncards );
	

	for( card=0; card<ncards; ++card ) {
		outval( 0 );	// Beginning of adapter marker
		config_one_card( card );
		outval( 0xff );	// End of adapter marker
	}

}

int main(int argc, char **argv)
{
	
	int argvval;
	
	cxconc_hlp=(char *)malloc(50);
	
#ifdef EPCXCONF
	printf( "epcxconf version %s\n", version );
#else
	printf( "cxconf version %s\n", version );
#endif
		
	if( (fout = fopen( FNAME, "w" )) == NULL ) {
		printf( "Cannot open '%s' for writing.\n", FNAME );
		exit( 1 );
	}
	
	if(argc == 1) {
		config(0);
	} else if(argc == 2) {
		if(!(argvval = atoi(argv[1]))) {
			printf("Invalid arguments. Halted\n");
			exit(0);
		} else {
			config(argvval);
		}
	} else if(argc > 2) {
		printf("Too many arguments. Dying!\n");
		exit(0);
	}
			
	fclose( fout );
	free(cxconc_hlp);
}
